#include <stdio.h> // printf(), fprintf(), perror()
#include <stdlib.h> // EXIT_FAILURE and EXIT_SUCCESS
#include <string.h> // strerror, memset
#include <stdint.h> // uint32_t and friends
#include <stdbool.h> // bool (in C, not C++)
#include <inttypes.h> // PRId32 and friends
#include <time.h> // clock_nanosleep(), clock_gettime()
#include <errno.h> // errno
// #include <signal.h> // sigaction

#include <sched.h> // cpu_set_t, CPU_SETSIZE, CPU_ISSET(), CPU_ZERO(), CPU_SET(), sched_param, sched_setscheduler(), sched_getaffinity() [with _GNU_SOURCE defined], sched_setaffinity()

//#include <sys/resource.h> // PRIO_PROCESS
//#include <sys/time.h> // timespec? (or in <time.h>?)
//#include <sys/types.h> // pid_t
#include <sys/mman.h> // mlockall()
#include <unistd.h> // getpid(), getuid()
#include <fcntl.h> // for open(), O_WRONLY, O_CLOEXEC

#include <ecrt.h>

// defines
#define PERIOD_NS   (1000000)
#define NSEC_PER_SEC (1000000000)
#define FREQUENCY (NSEC_PER_SEC / PERIOD_NS)

#define YASKAWA_SLAVE_VENDOR_CODE  ((uint32_t)0x00000539)
#define YASKAWA_SLAVE_PRODUCT_CODE ((uint32_t)0x02200301)

#define YASKAWA_SLAVE_ALIAS    (0)
#define YASKAWA_SLAVE_POSITION (0)

// #define DO_NONREALTIME_READ_TEST
#define USE_READ_DOMAIN
#define USE_WRITE_DOMAIN

// global variables
static ec_master_t *master = NULL;
static ec_domain_t *domain_r = NULL;
static ec_domain_t *domain_w = NULL;
static ec_slave_config_t *sc  = NULL;
#ifdef USE_READ_DOMAIN
static uint8_t *domain_r_pd = NULL;
#endif
#ifdef USE_WRITE_DOMAIN
static uint8_t *domain_w_pd = NULL;
#endif

// RxPDO entry offsets
static unsigned int ctrl_word;
static unsigned int target_pos;
static unsigned int tar_velo;
static unsigned int tar_torq;
static unsigned int max_torq;
static unsigned int modeofoper;
static unsigned int interpolateddata;

// TxPDO entry offsets
// static unsigned int ctrl_word_verify; // FOR TESTING
static unsigned int status_word;
static unsigned int actual_pos;
static unsigned int torq_actu_val;
static unsigned int following_actu_val;
static unsigned int modeofop_display;
static unsigned int touch_probe_stat;
static unsigned int touch_probe_val;

// RxPDO (1600h) has register CONTROLWORLD  6040h, target position (607Ah), etc.
// TXPDO (1A00h) has register STATUSWORD    6041h, position actual value (6064h), etc.

// RxPDO (1600h)
#ifdef USE_READ_DOMAIN
static const ec_pdo_entry_reg_t domain_r_regs[] =
{
        {YASKAWA_SLAVE_ALIAS,YASKAWA_SLAVE_POSITION,YASKAWA_SLAVE_VENDOR_CODE,YASKAWA_SLAVE_PRODUCT_CODE,0x6040,0,&ctrl_word       },
        {YASKAWA_SLAVE_ALIAS,YASKAWA_SLAVE_POSITION,YASKAWA_SLAVE_VENDOR_CODE,YASKAWA_SLAVE_PRODUCT_CODE,0x60FF,0,&target_pos      },
        // {YASKAWA_SLAVE_ALIAS,YASKAWA_SLAVE_POSITION,YASKAWA_SLAVE_VENDOR_CODE,YASKAWA_SLAVE_PRODUCT_CODE,0x60ff,0,&tar_velo        },
        // {YASKAWA_SLAVE_ALIAS,YASKAWA_SLAVE_POSITION,YASKAWA_SLAVE_VENDOR_CODE,YASKAWA_SLAVE_PRODUCT_CODE,0x6071,0,&tar_torq        },
        // {YASKAWA_SLAVE_ALIAS,YASKAWA_SLAVE_POSITION,YASKAWA_SLAVE_VENDOR_CODE,YASKAWA_SLAVE_PRODUCT_CODE,0x6072,0,&max_torq        },
        // {YASKAWA_SLAVE_ALIAS,YASKAWA_SLAVE_POSITION,YASKAWA_SLAVE_VENDOR_CODE,YASKAWA_SLAVE_PRODUCT_CODE,0x6060,0,&modeofoper      },
        // {YASKAWA_SLAVE_ALIAS,YASKAWA_SLAVE_POSITION,YASKAWA_SLAVE_VENDOR_CODE,YASKAWA_SLAVE_PRODUCT_CODE,0x60c1,1,&interpolateddata},
        {}
};
#endif

// TxPDO (1A00h)
#ifdef USE_WRITE_DOMAIN
static const ec_pdo_entry_reg_t domain_w_regs[] =
{
        {YASKAWA_SLAVE_ALIAS,YASKAWA_SLAVE_POSITION,YASKAWA_SLAVE_VENDOR_CODE,YASKAWA_SLAVE_PRODUCT_CODE,0x6041,0,&status_word       },
        {YASKAWA_SLAVE_ALIAS,YASKAWA_SLAVE_POSITION,YASKAWA_SLAVE_VENDOR_CODE,YASKAWA_SLAVE_PRODUCT_CODE,0x6064,0,&actual_pos        },
        // {YASKAWA_SLAVE_ALIAS,YASKAWA_SLAVE_POSITION,YASKAWA_SLAVE_VENDOR_CODE,YASKAWA_SLAVE_PRODUCT_CODE,0x6040,0,&ctrl_word_verify  },
        // {YASKAWA_SLAVE_ALIAS,YASKAWA_SLAVE_POSITION,YASKAWA_SLAVE_VENDOR_CODE,YASKAWA_SLAVE_PRODUCT_CODE,0x6077,0,&torq_actu_val     },
        // {YASKAWA_SLAVE_ALIAS,YASKAWA_SLAVE_POSITION,YASKAWA_SLAVE_VENDOR_CODE,YASKAWA_SLAVE_PRODUCT_CODE,0x60f4,0,&following_actu_val},
        // {YASKAWA_SLAVE_ALIAS,YASKAWA_SLAVE_POSITION,YASKAWA_SLAVE_VENDOR_CODE,YASKAWA_SLAVE_PRODUCT_CODE,0x6061,0,&modeofop_display  },
        // {YASKAWA_SLAVE_ALIAS,YASKAWA_SLAVE_POSITION,YASKAWA_SLAVE_VENDOR_CODE,YASKAWA_SLAVE_PRODUCT_CODE,0x60b9,0,&touch_probe_stat  },
        // {YASKAWA_SLAVE_ALIAS,YASKAWA_SLAVE_POSITION,YASKAWA_SLAVE_VENDOR_CODE,YASKAWA_SLAVE_PRODUCT_CODE,0x60ba,0,&touch_probe_val   },
        {}
};
#endif

/*
master idx="0" appTimePeriod="1000000" refClockSyncCycles="5">
   <slave idx="0" type="generic" vid="00000539" pid="02200301" configPdos="true">
         <dcConf assignActivate="300" sync0Cycle="*1" sync0Shift="0" />
         <syncManager idx="2" dir="out">
            <pdo idx="1600">
               <!-- <pdoEntry idx="6040" subIdx="00" bitLen="16" halPin="cia-controlword" halType="u32" />
               <pdoEntry idx="6060" subIdx="00" bitLen="8" halPin="opmode" halType="s32" />
               <pdoEntry idx="607A" subIdx="00" bitLen="32" halPin="target-position" halType="s32" />
               <pdoEntry idx="60FF" subIdx="00" bitLen="32" halPin="target-velocity" halType="s32" /> -->
            </pdo>
         </syncManager>
         <syncManager idx="3" dir="in">
            <pdo idx="1a00">
               <!-- <pdoEntry idx="6041" subIdx="00" bitLen="16" halPin="cia-statusword" halType="u32" />
               <pdoEntry idx="6061" subIdx="00" bitLen="8" halPin="opmode-display" halType="s32" />
               <pdoEntry idx="6064" subIdx="00" bitLen="32" halPin="actual-position" halType="s32" />
               <pdoEntry idx="606C" subIdx="00" bitLen="32" halPin="actual-velocity" halType="s32" />
               <pdoEntry idx="6077" subIdx="00" bitLen="32" halPin="actual-torque" halType="s32" /> -->
            </pdo>
         </syncManager>
   </slave>
</master>
*/

// PDO object indices, subindices, and data sizes: https://www.yaskawa.eu.com/Global%20Assets/Downloads/Technical_Documentation/Servo_Drives/Sigma7_Series/S7_Servopack/EtherCATs/Sigma-7_200V_AC_Servo_Drive_EtherCAT_Product_manual_SIEP_S800001_55.pdf#page=445
// https://www.yaskawa.eu.com/Global%20Assets/Downloads/Technical_Documentation/Servo_Drives/Sigma7_Series/S7_Servopack/EtherCATs/Sigma-7_200V_AC_Servo_Drive_EtherCAT_Product_manual_SIEP_S800001_55.pdf#page=486
// https://www.yaskawa.eu.com/Global%20Assets/Downloads/Technical_Documentation/Servo_Drives/Sigma7_Series/S7_Servopack/EtherCATs/Sigma-7_200V_AC_Servo_Drive_EtherCAT_Product_manual_SIEP_S800001_55.pdf#page=494
// https://www.yaskawa.eu.com/Global%20Assets/Downloads/Technical_Documentation/Servo_Drives/Sigma7_Series/S7_Servopack/EtherCATs/Sigma-7_200V_AC_Servo_Drive_EtherCAT_Product_manual_SIEP_S800001_55.pdf#page=498
// integer sizes: https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_plc_intro/2529399691.html&id=

#ifdef USE_READ_DOMAIN
static ec_pdo_entry_info_t domain_r_pdo_entry_infos[] = {
    {0x6040, 0x00, 16}, // control word
    {0x60FF, 0x00, 32}  // target velocity
};
#endif

#ifdef USE_WRITE_DOMAIN
static ec_pdo_entry_info_t domain_w_pdo_entry_infos[] = {
    {0x6041, 0x00, 16}, // status word
    {0x6064, 0x00, 32}, // position actual value
    // {0x6040, 0x00, 16}  // control word (VERIFY)
};
#endif

#ifdef USE_READ_DOMAIN
static ec_pdo_info_t yaskawa_r_pdos[] = {
    {0x1600, sizeof(domain_r_pdo_entry_infos)/sizeof(domain_r_pdo_entry_infos[0]), domain_r_pdo_entry_infos}
};
#endif

#ifdef USE_WRITE_DOMAIN
static ec_pdo_info_t yaskawa_w_pdos[] = {
    // {0x1A00}
    {0x1A00, sizeof(domain_w_pdo_entry_infos)/sizeof(domain_w_pdo_entry_infos[0]), domain_w_pdo_entry_infos}
};
#endif

static ec_sync_info_t yaskawa_syncs[] = {
    {0, EC_DIR_OUTPUT, 0, NULL, EC_WD_DISABLE},
    {1, EC_DIR_INPUT,  0, NULL, EC_WD_DISABLE},
#ifdef USE_READ_DOMAIN
    {2, EC_DIR_OUTPUT, sizeof(yaskawa_r_pdos)/sizeof(yaskawa_r_pdos[0]), yaskawa_r_pdos},
#else
    {2, EC_DIR_OUTPUT, 0, NULL, EC_WD_DISABLE},
#endif
#ifdef USE_WRITE_DOMAIN
    {3, EC_DIR_INPUT,  sizeof(yaskawa_w_pdos)/sizeof(yaskawa_w_pdos[0]), yaskawa_w_pdos},
#else
    {3, EC_DIR_INPUT,  0, NULL, EC_WD_DISABLE},
#endif
    {0xff}
};

static unsigned int counter = 0;

static void check_master_state(void)
{
    static ec_master_state_t old_ms;
    ec_master_state_t ms;

    ecrt_master_state(master, &ms);

    if (ms.slaves_responding != old_ms.slaves_responding) {
        printf("master slaves_responding: %u slave(s)\n", ms.slaves_responding);
    }
    if (ms.al_states != old_ms.al_states) {
        printf("master al_states: 0x%02X\n", ms.al_states);
    }
    if (ms.link_up != old_ms.link_up) {
        printf("master link_up: %s\n", ms.link_up ? "up" : "down");
    }

    old_ms = ms;
}

static void check_slave_config_states()
{
    static ec_slave_config_state_t old_s;
    ec_slave_config_state_t s;

    ecrt_slave_config_state(sc, &s);

    if (s.al_state != old_s.al_state) {
        printf("slave al_state 0x%02X\n", s.al_state);
    }
    if (s.online != old_s.online) {
        printf("slave online: %s\n", s.online ? "online" : "offline");
    }
    if (s.operational != old_s.operational) {
        printf("slave operational: %soperational\n", s.operational ? "" : "Not ");
    }

    old_s = s;
}

static void cyclic_task()
{
    // receive process data
    ecrt_master_receive(master);
#ifdef USE_READ_DOMAIN
    ecrt_domain_process(domain_r);
#endif
#ifdef USE_WRITE_DOMAIN
    ecrt_domain_process(domain_w);
#endif

    uint16_t VALUE_status_word;
    int32_t VALUE_actual_pos;
#ifdef USE_WRITE_DOMAIN
    VALUE_status_word = EC_READ_U16(domain_w_pd + status_word);
    VALUE_actual_pos = EC_READ_S32(domain_w_pd + actual_pos);
#endif

    if (counter)
    {
        counter--;
    }
    else
    {
        // do this at 4 Hz
        counter = FREQUENCY/4;
        check_master_state();
        check_slave_config_states();
        printf("actual pos: %10" PRId32 "; status word = 0x%08" PRIx16 "\n", VALUE_actual_pos, VALUE_status_word);
    }

    // TODO write process data

    // send process data
#ifdef USE_READ_DOMAIN
    ecrt_domain_queue(domain_r);
#endif
#ifdef USE_WRITE_DOMAIN
    ecrt_domain_queue(domain_w);
#endif
    ecrt_master_send(master);
}

//
/*void signal_handler(int signum) {
    switch (signum) {
        case SIGALRM:
            sig_alarms++;
            break;
    }
}*/

int main(int argc, char **argv)
{
    int cpu_dma_latency_fd = 0; // intentionally never closed, it will be cleaned up...
    const bool running_as_root = (getuid() == 0);

    if (running_as_root)
    {
        // set priority
        struct sched_param param;
        memset(&param, 0, sizeof(param));
        param.sched_priority = 90;
        if (sched_setscheduler(0, SCHED_FIFO, &param) == -1)
        {
            perror("sched_setscheduler failed");
        }

        // prevent memory from being paged out
        if (mlockall(MCL_CURRENT|MCL_FUTURE) == -1)
        {
            perror("mlockall failed");
        }

        // get the CPUs available to the process
        int rt_cpu_number = -1;
        {
            cpu_set_t cpuset_orig;
            if (sched_getaffinity(getpid(), sizeof(cpuset_orig), &cpuset_orig))
            {
                fprintf(stderr, "WARNING: sched_getaffinity() failed\n");
            }
            else
            {
                for (int i = 0; i < CPU_SETSIZE; i++)
                {
                    if (CPU_ISSET(i, &cpuset_orig))
                        rt_cpu_number = i;
                }
            }
        }

        if (rt_cpu_number != -1)
        {
            // set up a CPU set with only the desired CPU
            cpu_set_t cpuset;
            CPU_ZERO(&cpuset);
            CPU_SET(rt_cpu_number, &cpuset);

            // try to use the desired CPU
            const pid_t pid = getpid();
            if (sched_setaffinity(pid, sizeof(cpuset), &cpuset))
            {
                fprintf(stderr, "WARNING: setting CPU affinity failed\n");
            }
        }

        // prevent CPU from sleeping
        // open special device file for writing
        cpu_dma_latency_fd = open("/dev/cpu_dma_latency", O_WRONLY | O_CLOEXEC);
        if (cpu_dma_latency_fd < 0)
        {
            fprintf(stderr, "WARNING: failed to open /dev/cpu_dma_latency: %s\n", strerror(errno));
        }
        else
        {
            // write a 32-bit zero value
            static const uint32_t ZERO = 0x00000000;
            const ssize_t bytes_written = write(cpu_dma_latency_fd, &ZERO, sizeof(ZERO));
            if (bytes_written != sizeof(ZERO))
            {
                fprintf(stderr, "WARNING: failed to write to /dev/cpu_dma_latency: %s\n", strerror(errno));
            }
        }
    }

    master = ecrt_request_master(0);
    if (!master)
        return EXIT_FAILURE;

#ifdef DO_NONREALTIME_READ_TEST
    while (1)
    {
        uint8_t buffer[4];
        size_t buffer_len;
        uint32_t abort_code;
        // printf("calling ecrt_master_sdo_upload()...\n");
        if (ecrt_master_sdo_upload(master, YASKAWA_SLAVE_POSITION, 0x6064, 0, buffer, sizeof(buffer), &buffer_len, &abort_code) != 0)
        {
            fprintf(stderr, "ecrt_master_sdo_upload() returned an error code!\n");
            return -1;
        }
        printf("actual position value = %10" PRId32 "\n", EC_READ_S32(buffer));
        usleep(100*1000); // 100 milliseconds
    }
#endif

#ifdef USE_READ_DOMAIN
    domain_r = ecrt_master_create_domain(master);
    if (!domain_r)
        return EXIT_FAILURE;
#endif
#ifdef USE_WRITE_DOMAIN
    domain_w = ecrt_master_create_domain(master);
    if (!domain_w)
        return EXIT_FAILURE;
#endif

    if (!(sc = ecrt_master_slave_config(master, YASKAWA_SLAVE_ALIAS, YASKAWA_SLAVE_POSITION, YASKAWA_SLAVE_VENDOR_CODE, YASKAWA_SLAVE_PRODUCT_CODE)))
    {
        fprintf(stderr, "Failed to get slave1 configuration.\n");
        return EXIT_FAILURE;
    }

    ///////////

    // ecrt_slave_config_pdo_assign_clear(sc, 0);
    // ecrt_slave_config_pdo_assign_clear(sc, 1);
    // ecrt_slave_config_pdo_assign_clear(sc, 2);
    // ecrt_slave_config_pdo_assign_clear(sc, 3);
    // ecrt_slave_config_pdo_mapping_clear(sc, 0);
    // ecrt_slave_config_pdo_mapping_clear(sc, 1);
    // ecrt_slave_config_pdo_mapping_clear(sc, 2);
    // ecrt_slave_config_pdo_mapping_clear(sc, 3);

    if (ecrt_slave_config_pdos(sc, EC_END, yaskawa_syncs))
    {
        fprintf(stderr, "Failed to configure PDOs.\n");
        return EXIT_FAILURE;
    }

    ///////////

#ifdef USE_READ_DOMAIN
    if (ecrt_domain_reg_pdo_entry_list(domain_r, domain_r_regs))
    {
        fprintf(stderr, "read PDO entry registration failed!\n");
        return EXIT_FAILURE;
    }
#endif

#ifdef USE_WRITE_DOMAIN
    if (ecrt_domain_reg_pdo_entry_list(domain_w, domain_w_regs))
    {
        fprintf(stderr, "write PDO entry registration failed!\n");
        return EXIT_FAILURE;
    }
#endif

    // set distributed clock settings
    ecrt_slave_config_dc(sc, 0x300 /*assignActivate*/, PERIOD_NS, 0, 0 /*sync1cycle*/, 0);

    printf("Activating master...\n");
    if (ecrt_master_activate(master))
        return EXIT_FAILURE;

#ifdef USE_READ_DOMAIN
    if (!(domain_r_pd = ecrt_domain_data(domain_r)))
    {
        return EXIT_FAILURE;
    }
#endif

#ifdef USE_WRITE_DOMAIN
    if (!(domain_w_pd = ecrt_domain_data(domain_w)))
    {
        return EXIT_FAILURE;
    }
#endif

    /*pid_t pid = getpid();

    if (setpriority(PRIO_PROCESS, pid, -19))
        fprintf(stderr, "Warning: Failed to set priority: %s\n", strerror(errno));*/

    /*struct sigaction sa;
    sa.sa_handler = signal_handler;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    if (sigaction(SIGALRM, &sa, 0))
    {
        fprintf(stderr, "Failed to install signal handler!\n");
        return -1;
    }*/

    printf("Started.\n");

    struct timespec wakeup_time;
    clock_gettime(CLOCK_MONOTONIC, &wakeup_time);
    wakeup_time.tv_sec += 1; /* start in future */
    wakeup_time.tv_nsec = 0;

    while (1)
    {
        const int ret = clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &wakeup_time, NULL);
        if (ret)
        {
            fprintf(stderr, "clock_nanosleep(): %s\n", strerror(ret));
            return EXIT_FAILURE;
        }

        cyclic_task();

        wakeup_time.tv_nsec += PERIOD_NS;
        while (wakeup_time.tv_nsec >= NSEC_PER_SEC)
        {
            wakeup_time.tv_nsec -= NSEC_PER_SEC;
            wakeup_time.tv_sec++;
        }
    }

    return EXIT_SUCCESS;
}
